/* M-ACL algorithm for multiple predicate learning */
ver('1').

/*
Characteristics:

        Backtracking on clause addition

        Extensional coverage of example, using the original training set,
                not the only that is obtained by abductively extending it

        Test of the positive examples: first an (extensional only) prolog
        derivation for them is tried.
        If it fails, then the abductive derivation is tried.
        The extensional prolog derivation is implemented by means of the abd pp
        without abducibles.

        Non abducible are always inserted before literals in the body of rules, so
                that abducible always appear at the end

        Target clauses have the form
                clause((Head:-Body),Name,EplusCovered).

        The delta set has the form
                [[abd,list_of_clauses]]
        for each abducible, it is stored as well the list of clauses that
                generated it

M_ACL is based on the ACL1 algorithm for single predicate learning

Caracteristics:

        Top-down albgorithm with beam search

        Abductive proof procedure substitutes the prolog proof procedure

        Stopping criterion for the specialization loop: it stops when
                1) the first rule of the beam does not cover any negative ex
                2) a max number of specializing steps has been done

        No backtracking on clause addition

        optimization  of clause testing: the first resolution step
                is performed using only the current clause

        Hybrid extensional/intensional derivation: positive and negative examples
                are considered as facts that can be used in the derivation of examples
                for other predicates. All examples apart from the one we are
                        deriving are considered as facts in the derivation of the current
                example.

        Target predicates can be abducibles: assumptions about them are moved to the
                training set and become new training examples
*/

/*
Parameters:
        a number of parameters is defined by the facts at the beginning of the file.
        Set their value by changing the file and reloading it or by retract and asserts.

        verbosity(V): specifies the verbosity level.  Set by default to 3
        0: nothing is printed apart from the output rules
        1: every rule that is added to the theory
        2: every rule added to the agenda
        3: the agenda at each specializing step

        beamsize(Beamsize): specifies the size of the beam. Set by default to 5

        nmax(Nmax): maximum number of specializayion.  Set by default to 10

        der_depth(Depth): depth of derivations.  Set by default to 20

        kt(K):  threshold on K+ and K-.  Set by default to 0.1

	min_cov(MC): minimum number of examples that each clause must cover. 
		Set by default to 1

*/


kt(0.1). % k threshold

max_spec_steps(10).
der_depth(50).
beamsize(10).
verbosity(3).
%m(2).
min_cov(1).
stopping_acc(0.8).


/* Algorithm: (in parenthesis the name of the predicates)

induce: (i)
Theory:=0
Delta:=0
while E+!=0 do (covering_loop)
        initialize the agenda (initialize_agenda)
        generate one clause Cl (specialize)
        evaluate Cl in oder to find the examples E+(Cl) covered by Cl and the
                relative set of assumptions Delta(Cl)
        E+=E+ - E+(Cl)                                          \
        Add to E+ the pos lit of target pred in Delta(Cl)       |-(update)
        Add to E- the neg lit of target pred in Delta(Cl)       /
        Add Cl to the Theory
        Add Delta(Cl) to Delta
endwhile


specialize:
        pick the first clause in the agenda
                for each refinement (add_ref_to_agenda)
                        evaluate the refinement (evaluate)
                        if the refinement covers at least one pos ex
                        with or without abd
                                        orderly insert [Rule,Value] into Agenda
                                without exceeding Beamsize
                                (insert_in_order)
                endfor
        until the best rule in the agenda is consistent or the maximum number of
                specialization steps has been reached
        pick the first clause Cl in the agenda

Insertion of rules in the agenda:
        1) refinements with the same value of a rule already
                present are inserted after
        2) when 2 clauses have the same value for the heuristic,
                the clause with the higher number of pos ex covered without abd
                is put first


*/

:-consult(abdpp_mple).
:-consult(output).


i(File):-
        name(File,FileString),
        append(FileString,".bg",FileStringBg),
        name(FileBg,FileStringBg),
        [FileBg],
        append(FileString,".rules",FileOutString),
        name(FileOut,FileOutString),
        title(FileBg,FileOut,user_output),
        retract_old_pred,
        statistics(runtime,[_,_]),
        induce(Rules,_Delta),
        statistics(runtime,[_,T]),
        T1 is T /1000,
        open(FileOut,write,Stream),
        title(FileBg,FileOut,Stream),
        title(FileBg,FileOut,user_output),
        format(Stream,"/* Execution time ~f seconds. Generated rules */~N",[T1])
,
        format("/* Execution time ~f seconds. Generated rules */~N",[T1]),
        print_list(Rules,Stream),
        print_list(Rules,user_output),
        print_backtracked_clauses(user_output),
        print_backtracked_clauses(Stream),
        close(Stream).


induce(Rules,Delta):-
        eplus(Eplus),
        covering_loop(Eplus,[],[],[],Rules,[],Delta).

initialize_agenda(Agenda):-
        findall([H,V],bias(rule(H,_B),V,_Bias),L),
        init_ag(L,[],Agenda).

init_ag([],Agenda,Agenda):-!.

init_ag([[P,Var]|Rp],Agendain,[ [rule(P,[],Name),_V,_GC,_LC,_Np,_Npa,_Nm,_Nma,Var]
                | Agendaout]):-!,
        new_clause_name(Name),
        init_ag(Rp,Agendain,Agendaout).

/* covering_loop(Eplus,Eminus,Rulesin,Rulesout,Deltain,Deltaout) */
/* no more positive examples to cover */
covering_loop([],[],_EminusA,Rules,Rules,Delta,Delta):-!.

/* some eplus still to cover: generate a new clause */
covering_loop(Eplus,EplusA,EminusA,RulesIn,RulesOut,DeltaIn,DeltaOut):-
        print_ex_rem(Eplus,EplusA,EminusA),
        specialization_loop(Eplus,EplusA,EminusA,DeltaIn,
        Clause,Epc,Epac,Emac,Np,Npa,Nm,Nma,Delta1,GC,LC),
        deleteall(Delta1,DeltaIn,NewDelta),
        print_new_clause(Clause,GC,LC,Np,Npa,Nm,Nma,
                Epc,Epac,Emac,NewDelta),
        asserta_rule(Clause,Epc,Epac),
        update(Eplus,Epc,NEp1,EplusA,Epac,NEpa1,EminusA,NEma1,NewDelta),
        (GC=no->
                % backtracking on previous clauses
                backtracking(Emac,Delta1,Delta2,NEp1,NEp,NEpa1,NEpa,NEma1,NEma,RulesIn,Rules1)
        ;
                NEp=NEp1,
                NEpa=NEpa1,
                NEma=NEma1,

                Delta2=Delta1,
                Rules1=RulesIn
        ),!,
        covering_loop(NEp,NEpa,NEma,
                [[Clause,Np,Npa,Nm,Nma,Epc,Epac,Emac,NewDelta]
                |Rules1],RulesOut,Delta2,DeltaOut).

specialization_loop(Eplus,EplusA,EminusA,Deltain,
        Clause,Epc,Epac,Emac,Np,Npa,Nm,Nma,Delta1,GC,LC):-
        initialize_agenda(Agenda1),
        evaluate_new_agenda(Agenda1,[],Agenda,Eplus,EplusA,
                EminusA,Deltain),
        assert(loc_cons(no_loc_cons,0)),
        specialize(Agenda,Agenda2,Eplus,EplusA,EminusA,Deltain,0),
        (Agenda2\==[]->
        % globally consistent clause found
                Agenda2=[ [Clause,_V|_]|_Agenda3],
                retract(loc_cons(_Clause,_))
        ;
        % locally consisten clause found
                retract(loc_cons(Clause,_))
        ),
        evaluate(_Value,Clause,Eplus,Epc,EplusA,Epac,_Eminuscovered,
                EminusA,Emac,Deltain,Delta1,Np,Npa,Nm,Nma,GC,LC).


/* specialize(Agenda,Clout,AllowedLiterals,Eplus,Eminus,
        Deltain) */

/* stopping criterion (0): no globally consistent clause found, only locally
        consistent */
specialize([],[],
        _Eplus,_Epa,_EminusA,_Delta,_N):-!,
        loc_cons(Clause,_),
        Clause\==no_loc_con.

/* stopping criterion (1): when the 1st rule does not cover any negative
        example */
specialize([ [BestRule,Value,yes|RestRule] |RestAgenda],
        [ [BestRule,Value,yes|RestRule] |RestAgenda],
        _Eplus,_Epa,_EminusA,_Delta,_N):-!,
        print_current_agenda([ [BestRule,Value,yes|RestRule] |RestAgenda]).

/* stopping criterion (2): heuristic function > Threshold */
/* spcialize([ [BestRule,Value,_End] |_Rest],BestRule,_AllowedLiterals,_Eplus,
        _Eminus,_Delta,_N):-
        Value>0.95,!. */

/* stopping criterion (3): reached the limit on the number of specializing

        steps */
specialize(Agenda,Agenda,
        _Eplus,_Epa,_EminusA,_Delta,N):-
        max_spec_steps(Nmax),
        N=Nmax,
        print_max_spec_steps(Nmax),!.

specialize([ [BestRule,_Value,_GC,_LC,Np,Npa,Nm,Nma,Var] |Rest],Agendaout,
        Eplus,Epa,EminusA,Delta,N):-
        print_spec_step(N),
        print_current_agenda([ [BestRule,_Value,_GC,_LC,Np,Npa,Nm,Nma,Var]
                |Rest]),
        print_refinements,
        BestRule=rule(H,B,_Name),
        bias(rule(H,B),Var,AllowedLiterals),
        add_ref_to_agenda(BestRule,Var,AllowedLiterals,
                Eplus,Epa,EminusA,
                Delta, Rest, Agenda),
        N1 is N + 1,
        !,
        specialize(Agenda,Agendaout,Eplus,Epa,EminusA,Delta,N1).


add_ref_to_agenda(_Rule,_Var,[],_Eplus,_Epa,_EminusA,
        _Delta,Agenda,
        Agenda):-!.

add_ref_to_agenda(rule(H,B,Name),Var,[Lit|RestLit],Eplus,Epa,EminusA,Delta,
        Agendain,
        Agendaout):-
        \+(memberchk_eq(Lit,B)),!,
        (abducible(Lit)->
                append(B,[Lit],B1)
        ;
                separate_abd_nabd(B,[],Nabd,Abd),
                append(Nabd,[Lit],Nabd1),
                append(Nabd1,Abd,B1)
        ),
        new_clause_name(NewName),
        evaluate(Value,rule(H,B1,NewName),Eplus,_Epluscovered,Epa,_Epac,_Emc,EminusA,
                _EminuscoveredA,Delta,
                _Deltaout,Np,Npa,Nm,Nma,GC,LC),
        Npt is Np + Npa,
	min_cov(MC),
        (Npt>=MC->
                print_refinement([rule(H,B1,NewName),Value,GC,LC,Np,Npa,Nm,Nma]),
                beamsize(Beamsize),
                insert_in_order([rule(H,B1,NewName),Value,GC,LC,Np,Npa,Nm,Nma,Var],
                        Beamsize,Agendain,Agenda1),
	        (LC=yes->
        	        loc_cons(Clause,V),
                	(Value>V->
                        	retract(loc_cons(Clause,V)),
	
        	                assert(loc_cons(rule(H,B1,NewName),Value)),
                	        print_loc_cons(rule(H,B1,NewName),Value)
                	;
                        	true
                	)
        	;
                	true
        	)
        ;
                Agenda1=Agendain
        ),
        !, % tail rec. optim.
        add_ref_to_agenda(rule(H,B,Name),Var,RestLit,Eplus,Epa,EminusA,
        Delta,Agenda1,Agendaout).

/* case in which Lit is already in the body */
add_ref_to_agenda(rule(H,B,Name),Var,[_Lit|RestLit],Eplus,Epa,EminusA,
        Delta,Agendain,
        Agendaout):-!,
        add_ref_to_agenda(rule(H,B,Name),Var,RestLit,Eplus,Epa,EminusA,
                Delta,Agendain,
                Agendaout).


backtracking(Emac,DeltaIn,DeltaOut,EpIn,EpOut,EpaIn,EpaOut,EmaIn,EmaOut,
        RulesIn,RulesOut):-
        pick_clauses(Emac,DeltaIn,[],ClauseList),
        print_backtracking(ClauseList),
        remove_clause(ClauseList,DeltaIn,DeltaOut,EpIn,EpOut,EpaIn,EpaOut,EmaIn,EmaOut,RulesIn,RulesOut).



pick_clauses([],_D,Cl,Cl):-!.

pick_clauses([Ex|Rest],Delta,ClIn,ClOut):-
        findall(Clause, member([not(Ex),Clause],Delta),List),
        append(List,ClIn,Cl1),
        pick_clauses(Rest,Delta,Cl1,ClOut).


remove_clause([],Delta,Delta,Ep,Ep,Epa,Epa,Ema,Ema,Rules,Rules):-!.


remove_clause([Name|Rest],DeltaIn,DeltaOut,EpIn,EpOut,EpaIn,EpaOut,EmaIn,EmaOut,RulesIn,RulesOut):-
        retract(rule(H,B,Name,Epc,Epac)),
        assert(backtracked(rule(H,B,Name))),
        append(Epc,EpIn,Ep1),
        append(Epac,EpaIn,Epa1),
        remove_abduced(Name,DeltaIn,Delta1,Removed),
        remove_examples(Removed,Epa1,Epa2,EmaIn,Ema1),
        delete1(RulesIn,[rule(_,_,Name)|_],Rules1),
        remove_clause(Rest,Delta1,DeltaOut,Ep1,EpOut,Epa2,EpaOut,Ema1,EmaOut,
                Rules1,RulesOut).


remove_abduced(Name,DeltaIn,DeltaOut,RemovedEx):-
        findall(Ex,member([Ex,Name],DeltaIn),RemovedEx),
        findall([Ex,Name],member([Ex,Name],DeltaIn),RemovedDelta),
        deleteall(DeltaIn,RemovedDelta,DeltaOut).

remove_examples([],Epa,Epa,Ema,Ema):-!.

remove_examples([not(Ex)|Rest],EpaIn,EpaOut,EmaIn,EmaOut):-
        delete(EmaIn,Ex,Ema1),
        remove_examples(Rest,EpaIn,EpaOut,Ema1,EmaOut).

remove_examples([Ex|Rest],EpaIn,EpaOut,EmaIn,EmaOut):-
        % if the example was already covered, do not do anything
        (delete(EpaIn,Ex,Epa1);Epa1=EpaIn),
        remove_examples(Rest,Epa1,EpaOut,EmaIn,EmaOut).


/* ************** Predicates for clause evaluation ********* */
evaluate_new_agenda([],Agenda,Agenda,_,_,_,_):-!.

evaluate_new_agenda([[Rule,_V,GC,LC,Np,Npa,Nm,Nma,Var]|Agenda],
                AgendaNewin,
                AgendaNewout,
                Ep,Epa,Ema,Delta):-
        evaluate(Value,Rule,Ep,_Epc,Epa,_Epac,_Emc,Ema,_Emac,Delta,_Dout,Np,Npa,Nm,Nma,GC,LC),
        beamsize(Beamsize),
        Npt is Np+Npa,
        (Npt>0->
                insert_in_order([Rule,Value,GC,LC,Np,Npa,Nm,Nma,Var],
                                Beamsize,AgendaNewin,AgendaNew1)
        ;
                AgendaNew1=AgendaNewin
        ),
        evaluate_new_agenda(Agenda,AgendaNew1,AgendaNewout,Ep,Epa,Ema,Delta).




/*  evaluate(Value,rule(H,[]),Eplus,_Epluscovered,Eminus,_Eminuscovered,
                Delta,_Deltaout),
*/

evaluate(Value,Cl,Ep,Epc,Epa,Epac,Emc,
                Ema,Emac,Deltain,Deltaout,Np,Npa,Nm,Nma,GC,LC):-
        asserta_rule(Cl,[],[]),
        assert(Cl), %for optimization in the evaluation proc
        Cl=rule(H,_B,_N),
        functor(H,Hpred,_),
        eminus(Em),
        eplus(Eplus),
        append(Eplus,Epa,Ept),
        append(Em,Ema,Emt),
        asserta(current_eplus_tot(Ept)),
        asserta(current_eminus_tot(Emt)),
        epluscovered(Ep,[],Epc,Deltain,Delta1,0,Np,
                Hpred),
        eminuscovered(Em,[],Emc,Delta1,Delta2,0,Nm,
                Hpred),
        epluscovered(Epa,[],Epac,Delta2,Delta3,
                0,Npa,Hpred),
        eminuscovered(Ema,[],Emac,Delta3,Deltaout,
                0,Nma,Hpred),
        retract(Cl),
        retract_rule(Cl),
        retract(current_eplus_tot(Ept)),
        retract(current_eminus_tot(Emt)),
        (Nm=0->
                LC=yes,
                (Nma=0->
                        GC=yes
                ;
                        GC=no
                )
        ;
                LC=no,
                GC=no
        ),
        heuristic_function(Value,Np,Npa,Nm,Nma,Hpred).

heuristic_function(Value,Nplus,Nplusa,Nminus,Nminusa,_Hpred):-
        Value is (Nplus + Nplusa ) /
        (Nplus + Nminus + Nplusa + Nminusa ).

/* epluscovered(Eplus,Epluscoveredin,Epluscoveredout,Deltain,Deltaout,
                0,Nplus,0,Nplusa) */

epluscovered([],Epc,Epc,Delta,Delta,N,N,_Hpred):-!.
        % end of recursion: no more pos example to try

epluscovered([H|T],Epcin,Epcout,Deltain,Deltaout,Npin,Npout,Hp):-
        \+(functor(H,Hp,_)), % example not belonging to the current target pred
        epluscovered(T,Epcin,Epcout,Deltain,Deltaout,
                Npin,Npout,Hp).

epluscovered([H|T],Epcin,Epcout,Deltain,Deltaout,Npin,Npout,Hp):-
        assert_current_tr_set(H),
        der_depth(Depth),
        rule(H,Body,Name),        %optimization: the first resolution step is done

                        % with the rule
        add_cl_name(Body,Body1,Name),
        (coverage_test(H,Body1,Deltain,Depth,Delta1,yes)->
                % H is not covered abducing it
                Np1 is Npin +1,
                append(Delta1,Deltain,Delta2),
                remove_duplicates(Delta2,Delta3),
                retract_current_tr_set,
                epluscovered(T,[H|Epcin],Epcout,Delta3,Deltaout,Np1,Npout,
                        Hp)
        ;
                retract_current_tr_set,
                epluscovered(T,Epcin,Epcout,Deltain,Deltaout,Npin,Npout,
                        Hp)
        ).

coverage_test(_H,Body1,_Deltain,Depth,[],Ans):-
        retract(abducibles(LA)),
        asserta(abducibles([])),
        (solve_abd(Body1,[],[],Depth)->
                Ans1=yes
        ;
                Ans1=no
        ),
        retract(abducibles([])),
        asserta(abducibles(LA)),
        Ans1=Ans.
% in this way it fails only after having reasserted the abducibles set

coverage_test(H,Body1,Deltain,Depth,Delta1,yes):-
        solve_abd(Body1,[],Delta1,Depth),\+(member(H,Delta1)),
                \+ inconsistent(Delta1,Deltain).



inconsistent([[D,_]|_Rest],Deltain):-
        neg(D,NotD),
        member([NotD,_],Deltain),!.

inconsistent([_|Rest],Deltain):-
        inconsistent(Rest,Deltain).


eminuscovered([],Emc,Emc,Delta,Delta,N,N,_Hp):-!.
        % fine della ricorsione: non ci sono pi es neg da tentare

eminuscovered([H|T],Emcin,Emcout,Deltain,Deltaout,Nmin,Nmout,Hp):-
        \+(functor(H,Hp,_)), % example not belonging to the current target pred
        eminuscovered(T,Emcin,Emcout,Deltain,Deltaout,Nmin,Nmout,Hp).

eminuscovered([H|T],Emcin,Emcout,Deltain,Deltaout,Nmin,Nmout,Hp):-
        assert_current_tr_set(H),
        rule(H,Body,Name),
        add_cl_name(Body,Body1,Name),
        der_depth(Depth),
        ((solve_con2([Body1],[[not(H),top]],Delta1,Depth),
        \+ inconsistent(Delta1,Deltain))->
        % optimization: the rule is used immediately in the consistency phase.
        % hp: on not(H) there only the naf ic
                % ex not covered (not(H) succeeds)
                delete(Delta1,[not(H),top],Delta2),
                % not(H) removed from Delta
                append(Delta2,Deltain,Delta3),
                remove_duplicates(Delta3,Delta4),
                retract_current_tr_set,
                eminuscovered(T,Emcin,Emcout,Delta4,Deltaout,Nmin,Nmout,Hp)

        ;
                Nm1 is Nmin +1,
                retract_current_tr_set,
                eminuscovered(T,[H|Emcin],Emcout,Deltain,Deltaout,Nm1,Nmout,Hp)
        ).





/* ************* Auxiliary Predicates ***************** */

separate_abd_nabd([],Nabd,Nabd,[]):-!.

separate_abd_nabd([H|T],Nabd,Nabd,[H|T]):-
        abducible(H),!.

separate_abd_nabd([H|T],Nabdin,Nabdout,Abd):-
        append(Nabdin,[H],Nabd1),
        separate_abd_nabd(T,Nabd1,Nabdout,Abd).


update(Ep,Epc,NEp,Epa,Epac,NEpa,Ema,NEma,Delta1):-
        deleteall(Ep,Epc,NEp),
        deleteall(Epa,Epac,NEpa1),
        add_ex(Delta1,NEpa1,NEpa,Ema,NEma).

add_ex([],Eplus,Eplus,Eminus,Eminus):-!.

add_ex([[not(A),_]|DeltaRest],Eplusin,Eplusout,Eminusin,[A|Eminusout]):-
        !,
        add_ex(DeltaRest,Eplusin,Eplusout,Eminusin,Eminusout).

add_ex([[A,_]|DeltaRest],Eplusin,[A|Eplusout],Eminusin,Eminusout):-
        add_ex(DeltaRest,Eplusin,Eplusout,Eminusin,Eminusout).


insert_in_order(NewRuleItem,BeamSize,[],
        [NewRuleItem]):-
        BeamSize>0,!.

insert_in_order(_NewRuleItem,0,Agenda,Agenda):-!.

insert_in_order([NewRule,NewValue|NewRest],BeamSize,
        [ [Rule,Value|Rest] |RestAgendain],
        [ [NewRule,NewValue|NewRest],[Rule,Value|Rest] |RestAgendaout]):-
        (NewValue>Value % bigger heuristic
        ;
        NewValue=Value,
        NewRest=[_NGC,_NLC,NNp|_NR],
        Rest=[_GC,_LC,Np|_R],
        NNp>Np),!,  % equal heuristic, more pos ex covered
        length(RestAgendain,L),
        (L>BeamSize->
                nth(BeamSize,RestAgendain,_Last,RestAgendaout)
        ;
                RestAgendaout=RestAgendain
        ).

insert_in_order(NewRuleItem,BeamSize,
        [ RuleItem |RestAgenda],
        [ RuleItem |RestAgendaout]):-
        BeamSize1 is BeamSize -1,
        insert_in_order(NewRuleItem,BeamSize1,RestAgenda,
                RestAgendaout).

deleteallocc(H,Ep,Ep):-
        \+(member(H,Ep)),!.

deleteallocc(H,Epi,Epo):-
        functor(H,Hp,Ar),
        functor(H1,Hp,Ar),
        select(H1,Epi,Ep1),
        deleteallocc(H,Ep1,Epo).

% select a literal from the bias
select_literal_to_add(L,AllLitin,AllLitout):-
                        member(L,AllLitin),
                        delete(AllLitin,L,AllLitout).


new_clause_name(H):-
        retract(new_const_number_clauses(N)),
        N1 is N + 1,
        assert(new_const_number_clauses(N1)),
        number_chars(N,Nstr),
        append("c",Nstr,Nconst),
        name(H,Nconst).

:-dynamic new_const_number_clauses/1.

new_const_number_clauses(0).

/* ********* Predicates for modifying the database ********* */

retract_old_pred:-
        retractall(current_eplus(_)),
        retractall(current_eplus_tot(_)),
        retractall(current_eminus_tot(_)),
        retractall(current_eminus(_)),
        retractall(already_abduced(_)),
        retractall(rule(_,_,_)),
        retractall(rule(_,_,_,_,_)),
        retractall(loc_cons(_,_)),
        retractall(backtracked(_)).


assert_current_tr_set(Ex):-
        current_eplus_tot(Eplus),
        delete(Eplus,Ex,Epr),
        asserta(current_eplus(Epr)),
        current_eminus_tot(Eminus),
        delete(Eminus,Ex,Emr),
        asserta(current_eminus(Emr)).

retract_current_tr_set:-
        retract(current_eplus(_Epr)),
        retract(current_eminus(_Emr)).

asserta_rule(rule(H,B,Name),Epc,Epac):-
        assertz(rule(H,B,Name,Epc,Epac)).

retract_rule(rule(_H,_B,Name)):-
        retract(rule(_,_,Name,_,_)).

/* ********** Predicates for list handling *********** */

list2and([X],X):-
        \+(is_list(X)),!.

list2and([H|T],(H,Ta)):-!,
        list2and(T,Ta).

memberchk_eq(A, [A1|_]) :-A==A1, !.
memberchk_eq(A, [_|B]) :-
        memberchk_eq(A, B).

deleteall(Eplus,[],Eplus):-!.

deleteall(Eplusin,[H|T],Eplusout):-
        delete(Eplusin,H,Eplusout1),
        deleteall(Eplusout1,T,Eplusout).

delete1([],_El,[]):-!.

delete1([El|Tin],El,Tout):-!,
        delete1(Tin,El,Tout).

delete1([H|Tin],El,[H|Tout]):-!,
        delete1(Tin,El,Tout).
